/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

package mlterm;

public class Style {
  public int length = 0;
  public int start = 0;
  public int fg_color = 0;
  public int fg_pixel = 0;
  public int bg_color = 0;
  public int bg_pixel = 0;
  public boolean underline = false;
  public boolean strikeout = false;
  public boolean bold = false;
  public boolean italic = false;
}
