/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_ISO8859_CONV_H__
#define __EF_ISO8859_CONV_H__

#include "ef_conv.h"

ef_conv_t *ef_iso8859_1_conv_new(void);

ef_conv_t *ef_iso8859_2_conv_new(void);

ef_conv_t *ef_iso8859_3_conv_new(void);

ef_conv_t *ef_iso8859_4_conv_new(void);

ef_conv_t *ef_iso8859_5_conv_new(void);

ef_conv_t *ef_iso8859_6_conv_new(void);

ef_conv_t *ef_iso8859_7_conv_new(void);

ef_conv_t *ef_iso8859_8_conv_new(void);

ef_conv_t *ef_iso8859_9_conv_new(void);

ef_conv_t *ef_iso8859_10_conv_new(void);

/* a.k.a ISO8859-11 */
ef_conv_t *ef_tis620_2533_conv_new(void);

ef_conv_t *ef_iso8859_13_conv_new(void);

ef_conv_t *ef_iso8859_14_conv_new(void);

ef_conv_t *ef_iso8859_15_conv_new(void);

ef_conv_t *ef_iso8859_16_conv_new(void);

ef_conv_t *ef_tcvn5712_3_1993_conv_new(void);

#endif
