/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_8BIT_CONV_H__
#define __EF_8BIT_CONV_H__

#include "ef_conv.h"

ef_conv_t *ef_koi8_r_conv_new(void);

ef_conv_t *ef_koi8_u_conv_new(void);

ef_conv_t *ef_koi8_t_conv_new(void);

ef_conv_t *ef_georgian_ps_conv_new(void);

ef_conv_t *ef_cp1250_conv_new(void);

ef_conv_t *ef_cp1251_conv_new(void);

ef_conv_t *ef_cp1252_conv_new(void);

ef_conv_t *ef_cp1253_conv_new(void);

ef_conv_t *ef_cp1254_conv_new(void);

ef_conv_t *ef_cp1255_conv_new(void);

ef_conv_t *ef_cp1256_conv_new(void);

ef_conv_t *ef_cp1257_conv_new(void);

ef_conv_t *ef_cp1258_conv_new(void);

ef_conv_t *ef_cp874_conv_new(void);

ef_conv_t *ef_viscii_conv_new(void);

ef_conv_t *ef_tcvn5712_1_1993_conv_new(void);

ef_conv_t *ef_iscii_assamese_conv_new(void);

ef_conv_t *ef_iscii_bengali_conv_new(void);

ef_conv_t *ef_iscii_gujarati_conv_new(void);

ef_conv_t *ef_iscii_hindi_conv_new(void);

ef_conv_t *ef_iscii_kannada_conv_new(void);

ef_conv_t *ef_iscii_malayalam_conv_new(void);

ef_conv_t *ef_iscii_oriya_conv_new(void);

ef_conv_t *ef_iscii_punjabi_conv_new(void);

ef_conv_t *ef_iscii_tamil_conv_new(void);

ef_conv_t *ef_iscii_telugu_conv_new(void);

#endif
