/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_UCS_PROPERTY_H__
#define __EF_UCS_PROPERTY_H__

#include <pobl/bl_types.h> /* u_int32_t */

#include "ef_property.h"

ef_property_t ef_get_ucs_property(u_int32_t ucs);

#endif
