/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_ISO2022JP_PARSER_H__
#define __EF_ISO2022JP_PARSER_H__

#include "ef_parser.h"

ef_parser_t *ef_iso2022jp_7_parser_new(void);

ef_parser_t *ef_iso2022jp_8_parser_new(void);

ef_parser_t *ef_iso2022jp2_parser_new(void);

ef_parser_t *ef_iso2022jp3_parser_new(void);

#endif
