/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_ISO8859_PARSER_H__
#define __EF_ISO8859_PARSER_H__

#include "ef_parser.h"

ef_parser_t *ef_iso8859_1_parser_new(void);

ef_parser_t *ef_iso8859_2_parser_new(void);

ef_parser_t *ef_iso8859_3_parser_new(void);

ef_parser_t *ef_iso8859_4_parser_new(void);

ef_parser_t *ef_iso8859_5_parser_new(void);

ef_parser_t *ef_iso8859_6_parser_new(void);

ef_parser_t *ef_iso8859_7_parser_new(void);

ef_parser_t *ef_iso8859_8_parser_new(void);

ef_parser_t *ef_iso8859_9_parser_new(void);

ef_parser_t *ef_iso8859_10_parser_new(void);

/* a.k.a ISO8859-11 */
ef_parser_t *ef_tis620_2533_parser_new(void);

ef_parser_t *ef_iso8859_13_parser_new(void);

ef_parser_t *ef_iso8859_14_parser_new(void);

ef_parser_t *ef_iso8859_15_parser_new(void);

ef_parser_t *ef_iso8859_16_parser_new(void);

ef_parser_t *ef_tcvn5712_3_1993_parser_new(void);

#endif
