/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __EF_8BIT_PARSER_H__
#define __EF_8BIT_PARSER_H__

#include "ef_parser.h"

ef_parser_t *ef_koi8_r_parser_new(void);

ef_parser_t *ef_koi8_u_parser_new(void);

ef_parser_t *ef_koi8_t_parser_new(void);

ef_parser_t *ef_georgian_ps_parser_new(void);

ef_parser_t *ef_cp1250_parser_new(void);

ef_parser_t *ef_cp1251_parser_new(void);

ef_parser_t *ef_cp1252_parser_new(void);

ef_parser_t *ef_cp1253_parser_new(void);

ef_parser_t *ef_cp1254_parser_new(void);

ef_parser_t *ef_cp1255_parser_new(void);

ef_parser_t *ef_cp1256_parser_new(void);

ef_parser_t *ef_cp1257_parser_new(void);

ef_parser_t *ef_cp1258_parser_new(void);

ef_parser_t *ef_cp874_parser_new(void);

ef_parser_t *ef_viscii_parser_new(void);

ef_parser_t *ef_tcvn5712_1_1993_parser_new(void);

ef_parser_t *ef_iscii_assamese_parser_new(void);

ef_parser_t *ef_iscii_bengali_parser_new(void);

ef_parser_t *ef_iscii_gujarati_parser_new(void);

ef_parser_t *ef_iscii_hindi_parser_new(void);

ef_parser_t *ef_iscii_kannada_parser_new(void);

ef_parser_t *ef_iscii_malayalam_parser_new(void);

ef_parser_t *ef_iscii_oriya_parser_new(void);

ef_parser_t *ef_iscii_punjabi_parser_new(void);

ef_parser_t *ef_iscii_tamil_parser_new(void);

ef_parser_t *ef_iscii_telugu_parser_new(void);

#endif
