/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef __UI_DECSP_FONT_H__
#define __UI_DECSP_FONT_H__

#include "../ui.h"

int ui_load_decsp_xfont(XFontStruct *xfont, const char *decsp_id);

#endif
