/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#include <windows.h>

#define IDM_CONNECT    101
#define IDM_EXIT       102
#define IDM_EDIT       103
#define IDM_DEL        104
#define IDM_NEW_SERVER 105
#define IDM_NEW_FOLDER 106
#define IDM_SORT       107
#define IDD_FOLDER     110
#define IDD_SSH        120
#define IDD_MOSH       121
#define IDD_TELNET     122
#define IDD_RLOGIN     123
#define IDD_SERVER     124
#define IDD_PORT       125
#define IDD_USER       126
#define IDD_ENCODING   127
#define IDD_EXEC_CMD   128
#define IDD_SSH_PRIVKEY 129
#define IDD_X11        130

SM_MENU MENU DISCARDABLE
BEGIN
  POPUP "File(&F)"
  BEGIN
    MENUITEM "Connect(&C)", IDM_CONNECT
    MENUITEM "Exit(&X)", IDM_EXIT
  END
  POPUP "Edit(&E)"
  BEGIN
    MENUITEM "New Folder(&F)", IDM_NEW_FOLDER
    MENUITEM "New Server(&S)", IDM_NEW_SERVER
    MENUITEM "Edit(&E)", IDM_EDIT
    MENUITEM "Delete(&D)", IDM_DEL
    MENUITEM "Sort(&S)", IDM_SORT
  END
END

SM_FOLDER_DIALOG DIALOG DISCARDABLE 0, 0, 119, 40
STYLE DS_MODALFRAME | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "The name of the folder"
BEGIN
  DEFPUSHBUTTON "OK", IDOK, 7, 20, 50, 14
  PUSHBUTTON "Cancel",IDCANCEL, 59, 20, 50, 14
  EDITTEXT IDD_FOLDER, 7, 7, 105, 10, ES_AUTOHSCROLL
END

SM_EDIT_DIALOG DIALOG 20, 20, 151, 127
  STYLE WS_POPUP | WS_DLGFRAME | DS_CENTER
  {
    GROUPBOX "Protocol", -1, 4, 4, 143, 24
    RADIOBUTTON "&SSH", IDD_SSH, 8, 14, 28, 12
    RADIOBUTTON "&MOSH", IDD_MOSH, 36, 14, 35, 12
    RADIOBUTTON "&TELNET", IDD_TELNET, 71, 14, 35, 12
    RADIOBUTTON "&RLOGIN", IDD_RLOGIN, 106, 14, 35, 12
    LTEXT "Server", -1, 4, 34, 30, 8
    EDITTEXT IDD_SERVER, 38, 34, 96, 10, ES_AUTOHSCROLL
    LTEXT "Port", -1, 4, 45, 30, 8
    EDITTEXT IDD_PORT, 38, 45, 96, 10, ES_AUTOHSCROLL
    LTEXT "User", -1, 4, 56, 30, 8
    EDITTEXT IDD_USER, 38, 56, 96, 10, ES_AUTOHSCROLL
    LTEXT "Encoding", -1, 4, 67, 30, 8
    EDITTEXT IDD_ENCODING, 38, 67, 96, 10, ES_AUTOHSCROLL
    LTEXT "ExecCmd", -1, 4, 78, 30, 8
    EDITTEXT IDD_EXEC_CMD, 38, 78, 96, 10, ES_AUTOHSCROLL
    LTEXT "PrivKey", -1, 4, 89, 30, 8
    EDITTEXT IDD_SSH_PRIVKEY, 38, 89, 96, 10, ES_AUTOHSCROLL
    CONTROL "X11 forwarding", IDD_X11, "Button", BS_AUTOCHECKBOX | WS_GROUP | WS_TABSTOP, 4, 100, 126, 8
    DEFPUSHBUTTON "OK", IDOK, 20, 112, 40, 12
    PUSHBUTTON "Cancel", IDCANCEL, 80, 112, 40, 12
  }
