/* -*- c-basic-offset:2; tab-width:2; indent-tabs-mode:nil -*- */

#ifndef ___UI_CONNECT_DIALOG_H__
#define ___UI_CONNECT_DIALOG_H__

#include "../ui_connect_dialog.h"

/* Exported for winrs.rs. */
#define IDD_SSH 10
#define IDD_MOSH 11
#define IDD_TELNET 12
#define IDD_RLOGIN 13
#define IDD_SERVER 14
#define IDD_PORT 15
#define IDD_USER 16
#define IDD_PASS 17
#define IDD_ENCODING 18
#define IDD_EXEC_CMD 19
#define IDD_SSH_PRIVKEY 20
#define IDD_X11 21

#endif
